/*
Copyright (C) 1999/2000 Hendrik Lipka.
All rights reserved.

This program is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License along with this
program; if not, write to the Free Software Foundation, Inc., 675 Mass Ave,
Cambridge, MA 02139, USA or see http://www.gnu.org

If you encounter problems or have a suggestion, please mail to:
q3master@hendriklipka.de
*/

package de.hendriklipka.q3master;

/**
 * Title:        Quake3Arena MasterBrowser
 * Description:
 * Copyright:    Copyright (c) 2001 by Hendrik Lipka
 * @author Hendrik Lipka
 * @version 0.1
*/

import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;

import java.util.StringTokenizer;
import java.util.NoSuchElementException;

import org.log4j.*;

public class RequestHandler extends Thread
{
    private Category CAT=Category.getInstance(RequestHandler.class.getName());
    private Category LOG_CAT=Category.getInstance("q3master.request");

    protected final static String REQ_STR_HEARTBEAT="heartbeat QuakeArena";
    protected final static String REQ_STR_GETSERVERS="getservers";

    // we can use instance variables, because we create a new class instance per request
    protected byte[] _request=null;
    protected InetAddress _address=null;
    protected int _port=-1;
    protected DatagramSocket _socket=null;
    protected StatusHandler _handler=null;

    public RequestHandler(byte[] requestData, InetAddress address, int port, DatagramSocket socket, StatusHandler handler)
    {
        _handler=handler;
        _request=requestData;
        _address=address;
        _port=port;
        _socket=socket;
    }

    public void run()
    {
        /* dumping request
        for (int i=0;i<_request.length;i++)
        {
            System.out.println("#"+i+":"+_request[i]+"="+(char)_request[i]);
        }
        */
        // request should be at least 16 bytes
        if (_request.length<16)
        {
            LOG_CAT.error("request too short ("+_request.length+" bytes)");
            return;
        }
        // requests start with 4 bytes 0xFF
        for (int i=0;i<4;i++)
        {
            if (((byte)0xff)!=_request[i])
            {
                LOG_CAT.error("byte #"+i+" should be 0xff, but it is "+_request[i]);
                return;
            }
        }
        // check if heartbeat
        if (_request.length>(4+REQ_STR_HEARTBEAT.length()))
        {
            boolean found=true;
            for (int i=4;i<(4+REQ_STR_HEARTBEAT.length());i++)
            {
                byte b=_request[i];
                char c=REQ_STR_HEARTBEAT.charAt(i-4);
                if (b!=c)
                {
                    found=false;
                    break;
                }
            }
            if (found)
            {
//                System.out.println("server type="+(char)_request[_request.length-2]);
                LOG_CAT.info("got a heartbeat from "+_address.getHostAddress());
                handleHeartBeat();
                return;
            }
        }

        // check if getserver request
        if (_request.length>(4+REQ_STR_GETSERVERS.length()))
        {
            boolean found=true;
            for (int i=4;i<(4+REQ_STR_GETSERVERS.length());i++)
            {
                byte b=_request[i];
                char c=REQ_STR_GETSERVERS.charAt(i-4);
                if (b!=c)
                {
                    found=false;
                    break;
                }
            }
            if (found)
            {
                handleGetServer();
                return;
            }
        }
        LOG_CAT.error("error: unkown request!");
    }

    /**
     */
    public void handleHeartBeat()
    {
        // create server data object and set known data
        ServerData sd=new ServerData();
        sd._serverAddress=_address;
        sd._serverPort=_port;
        sd._lastHeartBeat=System.currentTimeMillis();
        // check for localservers (127.*)
        // they need to be replaced with their non-local IP address
        if (_address.getAddress()[0]==127)
        {
            try
            {
                CAT.debug("loopback found - trying alternate addresses for "+InetAddress.getLocalHost().getHostName());
                if (InetAddress.getLocalHost().getAddress()[0]!=127)
                {
                    CAT.debug("using local address directly - it's not loopback");
                    _address=InetAddress.getLocalHost();
                    sd._serverAddress=_address;
                }
                else
                {
                    InetAddress[] iaa=InetAddress.getAllByName(_address.getHostName());
                    CAT.info("found "+iaa.length+" addresses");
                    for (int i=0;i<iaa.length;i++)
                    {
                        CAT.info("alternate address: "+iaa[i].getHostAddress());
                        if (iaa[i].getAddress()[0]==127)
                            continue;
                        CAT.debug("setting alternate address");
                        _address=iaa[i];
                        sd._serverAddress=_address;
                        break;
                    }
                }
            }
            catch (Exception ex)
            {} // ignore - if we cannot resolve our local address, we have bigger problems...
        }
        // send request to server and set missing data
        _handler.addServer(sd);
    }

    /**
     *
     */
    public void handleGetServer()
    {
        int protocol=-1;
        boolean full=false;
        boolean empty=false;
        String s=new String(_request,REQ_STR_GETSERVERS.length()+5,_request.length-(REQ_STR_GETSERVERS.length()+5));
        LOG_CAT.info("got a getServers request: <"+s+">");
        StringTokenizer st=new StringTokenizer(s," ");
        try
        {
            String prot=(String)st.nextElement();
            protocol=Integer.parseInt(prot);
            String first=(String)st.nextElement();
            if (first.equals("empty"))
                empty=true;
            if (first.equals("full"))
                full=true;
            // max 2 specifiers for server filtering
            // @todo use unlimited options - maybe other mods use more values...
            first=(String)st.nextElement();
            if (first.equals("empty"))
                empty=true;
            if (first.equals("full"))
                full=true;
        }
        // ignore all exceptions - they can happen...
        catch (NoSuchElementException ex)
        {
        }
        catch (NumberFormatException ex)
        {
        }
        CAT.debug("protocol="+protocol);
        if (full)
            CAT.debug("include full servers");
        if (empty)
           CAT.debug("include empty servers");
        ServerData[] sda=_handler.getAllServers(protocol,full,empty);
        if (null==sda)
        {
            CAT.debug("got no servers from handler");
            return;
        }
        CAT.info("got "+sda.length+" servers from handler");
        ProtocolVersionHandler pvh=ProtocolVersionHandlerFactory.getProtocolVersionHandler(protocol);
        byte[][] responses=pvh.createResponse(sda);
        if (null==responses)
        {
            LOG_CAT.error("error: ProtocolHandler doesn't create response :(");
            return;
        }
        for (int i=0;i<responses.length;i++)
        {
            byte[] response=responses[i];
            if (null==response)
            {
                LOG_CAT.error("error: ProtocolHandler has created empty packets :(");
                continue;
            }

            try
            {
                // dump - for debugging purposes
                /*
                for (int j=0;j<response.length;j++)
                {
                    int ib=(response[j]<0)?(256+response[j]):(response[j]);
                    System.out.println("#"+j+":"+ib+"="+(char)response[j]);
                }
                */
                DatagramPacket packet=new DatagramPacket(response,response.length,_address,_port);
                CAT.debug("sending reponse ("+response.length+" bytes) to "+packet.getAddress().getHostAddress()+":"+packet.getPort());
                Object o=new Object();
                _socket.send(packet);
            }
            catch (Exception ex)
            {
                ex.printStackTrace();
            }
        }
    }
}